from .sql_types import (
    VARCHAR,
    CHAR,
    BLOB,
    TINYBLOB,
    TINYTEXT,
    MEDIUMBLOB,
    MEDIUMTEXT,
    LONGBLOB,
    LONGTEXT,
    ENUM,
    TEXT,
    STR,
    BOOL,
    INT,
    TINYINT,
    SMALLINT,
    MEDIUMINT,
    BIGINT,
    DECIMAL,
    FLOAT,
    DOUBLE,
    DATE,
    DATETIME,
    TIMESTAMP,
    TIME,
    YEAR
    )


type_map = {
    'VARCHAR':VARCHAR,
    'CHAR':CHAR,
    'BLOB':BLOB,
    'TINYBLOB':TINYBLOB,
    'TINYTEXT':TINYTEXT,
    'MEDIUMBLOB':MEDIUMBLOB,
    'MEDIUMTEXT':MEDIUMTEXT,
    'LONGBLOB':LONGBLOB,
    'LONGTEXT':LONGTEXT,
    'ENUM':ENUM,
    'TEXT':TEXT,
    'STR':STR,
    'BOOL':BOOL,
    'INT':INT,
    'TINYINT':TINYINT,
    'SMALLINT':SMALLINT,
    'MEDIUMINT':MEDIUMINT,
    'BIGINT':BIGINT,
    'DECIMAL':DECIMAL,
    'FLOAT':FLOAT,
    'DOUBLE':DOUBLE,
    'DATE':DATE,
    'DATETIME':DATETIME,
    'TIMESTAMP':TIMESTAMP,
    'TIME':TIME,
    'YEAR':YEAR,
    }

__all__ = (
    'VARCHAR',
    'CHAR',
    'BLOB',
    'TINYBLOB',
    'TINYTEXT',
    'MEDIUMBLOB',
    'MEDIUMTEXT',
    'LONGBLOB',
    'LONGTEXT',
    'ENUM',
    'TEXT',
    'STR',
    'BOOL',
    'INT',
    'TINYINT',
    'SMALLINT',
    'MEDIUMINT',
    'BIGINT',
    'DECIMAL',
    'FLOAT',
    'DOUBLE',
    'DATE',
    'DATETIME',
    'TIMESTAMP',
    'TIME',
    'YEAR',
    )


import datetime as dt
from sqlalchemy import Table, MetaData


class TableCache:
    def __init__(self, database:str, table_name:str, table):
        self.database   = database
        self.table_name = table_name
        self.table      = table
        self.expires_at = self.now+dt.timedelta(minutes=30)
        
    @classmethod
    def new(cls, database:str, table_name:str, engine):
        metadata = MetaData()
        table = Table(table_name, metadata, autoload_with=engine)
        return cls(database, table_name, table)
        
    def reset(self, engine):
        metadata = MetaData()
        self.table = Table(self.table_name, metadata, autoload_with=engine)
        self.expires_at = self.now+dt.timedelta(minutes=15)
        
    @property
    def now(self):
        return dt.datetime.now()
    
    @property
    def expired(self):
        n=self.now;e=self.expires_at
        return n >= e
            
    def __repr__(self):
        return f"<TableCache (`{self.database}`.`{self.table_name}`)>"


import datetime as dt
from sqlalchemy import MetaData


class TableCache:
    def __init__(self, database:str, table_name:str, table):
        self.database   = database
        self.table_name = table_name
        self.table      = table
        self.expires_at = self.now+dt.timedelta(minutes=30)
        
    @classmethod
    async def new(cls, database:str, table_name:str, engine):
        metadata = MetaData()
        async with engine.begin() as conn:
            # Reflect the table
            await conn.run_sync(metadata.reflect, only=[table_name])
        # Access the reflected Table object
        table = metadata.tables[table_name]
        return cls(database, table_name, table)
        
    async def reset(self, engine):
        metadata = MetaData()
        async with engine.begin() as conn:
            await conn.run_sync(metadata.reflect, only=[self.table_name])
        table = metadata.tables[self.table_name]
        self.table = table
        self.expires_at = self.now+dt.timedelta(minutes=15)
        
    @property
    def now(self):
        return dt.datetime.now()
    
    @property
    def expired(self):
        n=self.now;e=self.expires_at
        return n >= e
            
    def __repr__(self):
        return f"<TableCache (`{self.database}`.`{self.table_name}`)>"


# -*- coding: utf-8 -*-
from decimal import Decimal
from sqlalchemy.dialects import mysql

class DataBaseResponse():
    def __init__(self,
                 query,
                 as_decimals:bool=True,
                 response_model:object=None,
                 **dbr_args
                 ):
        # defaults
        self.database = None
        self.query    = query
        if query is None:
            self.raw_query = None
        else:
            self.raw_query = str(query.compile(
                dialect=mysql.dialect(),
                compile_kwargs={"literal_binds": False})
                )
        self.as_decimals = as_decimals
        self.response_model = response_model # callable or class that takes one positional argument 'row'
        # some default values
        self.keys         = []
        self.returns_rows = False
        self.records      = []
        self.successful   = None
        self.error_info   = None
        
        # check if any dbr args were passed that haven't been configured yet
        if len(dbr_args) > 0:
            keys = '\n'.join(f" - {key}" for key in dbr_args)
            print(f"WARNING!\nTHE FOLLOWING DataBaseResponse CONFIGURATION ARGUMENTS HAVEN'T BEEN CONFIGURED YET: \n{keys}")
        
    @classmethod
    def default(cls, database:str):
        dbr = cls(None)
        dbr.database = database
        return dbr
        
    async def _receive(self, result):
        try:
            self.successful = True
            self.returns_rows = result.returns_rows
            if result.returns_rows:
                self.keys = list(result.keys())
                data = result.fetchall()
                self.records = list(self.build_records(data))
        except Exception as e:
            self.successful = False
            self.error_info = str(e.__repr__())
            
    def build_records(self, data):
        while len(data) > 0:
            x = data.pop(0)
            yield self.build_record(x)
            
    def build_record(self, x):
        row = dict(zip(self.keys, x))
        if not self.as_decimals:
            for k in row:
                if isinstance(row[k], Decimal):
                    row[k] = float(row[k])
        # see if we need to return models or just dictionaries
        if self.response_model is not None:
            maker = self.response_model
        else:
            maker = lambda row: row
        row = maker(row)
        return row
    
    @property
    def row_count(self):
        return len(self)
    
    def __getitem__(self, k:int):
        return self.records[k]
    
    def __iter__(self):
        for x in self.records:
            yield x
            
    def __len__(self):
        return len(self.records)
    
    def __in__(self, value):
        return value in self.records
    
    def pop(self, index:int=-1):
        itm = self.records.pop(index)
        return itm
    
    def __repr__(self):
        return f"<DataBase Response ({len(self)} Records)>"


# -*- coding: utf-8 -*-
import datetime as dt
import json
from dbmasta.exceptions import InvalidDate

def unpack(obj, **kwargs):
    for k,v in kwargs.items():
        t = getattr(obj, k, 'NON-EXISTENT-VALUE')
        if t == 'NON-EXISTENT-VALUE':
            setattr(obj, k, v)

ESCAPE_CHARS = [
    # ('\\','\\\\'), ('%','%%'), ("'","''")
    ]

### TEXT
class VARCHAR:
    _type = str
    def __init__(self, 
                 value,
                 length = 255,
                 **kwargs):
        self._value = value
        self._length = kwargs.get('CHARACTER_MAXIMUM_LENGTH',length)
        self.nullable = kwargs.get('IS_NULLABLE', False)
        unpack(self, **kwargs)
        
    def clearEscapeChars(self):
        v = str(self._value)
        for x,y in ESCAPE_CHARS:
            v = v.replace(x,y)
        return v
        
    @property
    def value(self):
        if self._value is None:
            return None
        else:
            if isinstance(self._value,dict) or isinstance(self._value, list):
                v = json.dumps(self._value, default=str)
            v = self.clearEscapeChars()
            v = v[:min(len(v),self._length)]
            return v
    
    @property
    def SQL(self):
        if self.value is None and self.nullable:
            return 'NULL'
        return f"'{self.value}'"
    
    def __repr__(self):
        return f"'{self.value}'"


class CHAR:
    _type = str
    def __init__(self, 
                 value:str,
                 length = 255,
                 **kwargs
                 ):
        self._value = value
        self._length = kwargs.get('CHARACTER_MAXIMUM_LENGTH',length)
        self.nullable = kwargs.get('IS_NULLABLE', False)
        unpack(self, **kwargs)
        
    @property
    def value(self):
        if self._value is None:
            return None
        else:
            return str(self._value)[:min(len(str(self._value)),self._length)]
    
    @property
    def SQL(self):
        if self.value is None and self.nullable:
            return 'NULL'
        return f"'{self.value}'"
    
    def __repr__(self):
        return self.value


class STR(VARCHAR):
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)


class BLOB(VARCHAR): # CASE SENSITIVE
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)
        
    @property
    def value(self):
        if isinstance(self._value, dict):
            return json.dumps(self._value, default=str)
        else:
            return self._value

class TEXT(VARCHAR):
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)


class TINYBLOB(VARCHAR): # CASE SENSITIVE
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)
        
    @property
    def value(self):
        if isinstance(self._value, dict):
            return json.dumps(self._value, default=str)
        else:
            return self._value

class TINYTEXT(VARCHAR):
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)


class MEDIUMBLOB(BLOB): # CASE SENSITIVE
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)
        
    @property
    def value(self):
        if isinstance(self._value, dict):
            return json.dumps(self._value, default=str)
        else:
            return self._value

class MEDIUMTEXT(VARCHAR):
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)


class LONGBLOB(BLOB): # CASE SENSITIVE
    def __init__(self, value, **kwargs):
        super().__init__(value=value, **kwargs)
        
    @property
    def value(self):
        if isinstance(self._value, dict):
            return json.dumps(self._value, default=str)
        else:
            return self._value

class LONGTEXT(VARCHAR):
    def __init__(self, value, **kwargs):
        super().__init__(value=value, length=4_294_967_295, **kwargs)

### 
class ENUM(VARCHAR):
    def __init__(self, value:str, **kwargs):
        super().__init__(value=value,length=65535, **kwargs)
        # TODO include validator to make sure value is in the enum list





### INTEGERS
class INT:
    _type = int
    def __init__(self, value, **kwargs):
        self._value = value
        unpack(self, **kwargs)
        
    @property
    def value(self):
        if not isinstance(self._value, self._type):
            if str(self._value).isnumeric():
                return int(round(float(self._value),0))
            else:
                return 0
        else:
            return self._value
    
    @property
    def SQL(self):
        return f"{self.value}"
    
    def __repr__(self):
        return f"{self.value}"

class TINYINT(INT):
    pass

class SMALLINT(INT):
    pass

class MEDIUMINT(INT):
    pass

class BIGINT(INT):
    pass

class BOOL(TINYINT):
    def __init__(self, value, **kwargs):
        if value:
            value = 1
        else:
            value = 0
        super().__init__(value=value)


### FLOATS
class FLOAT:
    _type = float
    def __init__(self, value, **kwargs):
        self._value = value
        unpack(self, **kwargs)
        
    @property
    def value(self):
        test = str(self._value).replace('(','-').replace(')','').replace(',','').replace(' ','')
        accept = ['0','1','2','3','4','5','6','7','8','9','-','.']
        if not isinstance(self._value, self._type):
            if all([str(x) in accept for x in list(test)]):
                return float(test or 0)
            else:
                return 0
        else:
            return self._value
    
    @property
    def SQL(self):
        return f"{self.value}"
    
    def __repr__(self):
        return f"{self.value}"

class DECIMAL(FLOAT):
    pass

class DOUBLE(FLOAT):
    pass


### DATES & TIMES
class DATE:
    _type = dt.date
    def __init__(self, value, **kwargs):
        self._value = value
        unpack(self, **kwargs)
        _=self.value # force check type
    
    @property
    def value(self):
        if isinstance(self._value, dt.datetime):
            return self._value.date()
        elif isinstance(self._value, self._type):
            return self._value
        elif ((self._value == '' or self._value is None or self._value == 'NULL') and 
              getattr(self, 'IS_NULLABLE', True)):
            return None # 'NULL'
        else:
            raise InvalidDate(f'"{self._value}" must be of type datetime.date')
    
    @property
    def SQL(self):
        if self.value == 'NULL':
            return self.value
        else:
            return f"'{self.value}'"
    
    def __repr__(self):
        return str(self.value)

class DATETIME:
    _type = dt.datetime
    def __init__(self, value, **kwargs):
        self._value = value
        unpack(self, **kwargs)
        _=self.value # force check type
    
    @property
    def value(self):
        if isinstance(self._value, self._type):
            v = self._value
            v = v.replace(microsecond=0)
            return v
        elif ((self._value == '' or self._value == None or self._value == 'NULL') and 
              getattr(self, 'IS_NULLABLE', True)):
            return None
        else:
            raise InvalidDate(f'"{self._value}" must be of type datetime.datetime')
    
    @property
    def SQL(self):
        if self.value == 'NULL':
            return self.value
        else:
            return f"'{self.value}'"
    
    def __repr__(self):
        return str(self.value)

class TIMESTAMP:
    _type = dt.datetime
    def __init__(self, value, **kwargs):
        self._value = value
        unpack(self, **kwargs)
        _=self.value # force check type
    
    @property
    def value(self):
        if isinstance(self._value, self._type):
            v = self._value
            v = v.replace(microsecond=0)
            v = str(v)
            # v = v.replace('-','').replace(' ','').replace(':','')
            return v
        elif ((self._value == '' or self._value == None or self._value == 'NULL') and 
              getattr(self, 'IS_NULLABLE', True)):
            return None
        else:
            raise InvalidDate(f'"{self._value}" must be of type datetime.datetime')
    
    @property
    def SQL(self):
        if self.value == 'NULL':
            return self.value
        else:
            return f"'{self.value}'"
    
    def __repr__(self):
        return str(self.value)

class TIME:
    _type = dt.datetime
    def __init__(self, value, **kwargs):
        self._value = value
        unpack(self, **kwargs)
        _=self.value # force check type
    
    @property
    def value(self):
        if isinstance(self._value, self._type):
            v = self._value
            v = v.replace(microsecond=0)
            v = str(v).split(' ')[1]
            return v
        elif ((self._value == '' or self._value == None or self._value == 'NULL') and 
              getattr(self, 'IS_NULLABLE', True)):
            return None
        else:
            raise InvalidDate(f'"{self._value}" must be of type datetime.datetime')
    
    @property
    def SQL(self):
        if self.value == 'NULL':
            return self.value
        else:
            return f"'{self.value}'"
    
    def __repr__(self):
        return str(self.value)

class YEAR:
    _type = dt.datetime
    def __init__(self, value, **kwargs):
        self._value = value
        unpack(self, **kwargs)
        _=self.value # force check type
    
    @property
    def value(self):
        if isinstance(self._value, self._type):
            v = self._value
            return v.year
        elif ((self._value == '' or self._value == None or self._value == 'NULL') and 
              getattr(self, 'IS_NULLABLE', True)):
            return None
        else:
            raise InvalidDate(f'"{self._value}" must be of type datetime.datetime')
    
    @property
    def SQL(self):
        if self.value == 'NULL':
            return self.value
        else:
            return f"'{self.value}'"
    
    def __repr__(self):
        return str(self.value)



from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy import (text as sql_text, asc, desc, null as sql_null)
from sqlalchemy.sql import (and_, or_, 
                            not_ as sql_not, 
                            select, 
                            #insert, 
                            update as sql_update, 
                            delete,
                            join as sql_join
                            )
from sqlalchemy.dialects.mysql import insert
from sqlalchemy.pool import AsyncAdaptedQueuePool
import datetime as dt
from dbmasta.authorization import Authorization
from dbmasta.async_db_client.response import DataBaseResponse
from dbmasta.sql_types import type_map
from dbmasta.async_db_client.tables import TableCache

DEBUG = False
TIMEOUT_SECONDS = 240

class AsyncDataBase():
    Authorization = Authorization
    BLDRS = {
        'select': select,
        'insert': insert,
        'update': sql_update,
        'delete': delete,
        'text': sql_text,
        'and': and_,
        'or': or_,
        'not': sql_not,
        'asc': asc,
        'desc': desc,
        'null': sql_null,
        'join': sql_join
    }
    def __init__(self,
                 auth:Authorization,
                 ):
        self.auth = auth
        self.debug        = DEBUG
        self.table_cache  = {}
        
    @property
    def database(self):
        return self.auth.default_database
        
    @classmethod
    def env(cls):
        auth = Authorization.env(as_async=True)
        return cls(auth)

    def engine(self, database):
        return create_async_engine(
            url = self.auth.uri(database),
            echo = self.debug,
            poolclass = AsyncAdaptedQueuePool,  # Use a AsyncAdaptedQueuePool for pooling connections
            max_overflow = 0,       # No extra connections beyond the pool size
            pool_size = 1,          # Pool size of 1, mimicking single connection behavior
            pool_recycle = -1,      # Disables connection recycling
            pool_timeout = 30,      # Timeout for getting a connection from the pool
            connect_args = {'connect_timeout': TIMEOUT_SECONDS})

    async def preload_tables(self, db_tbls:list[tuple[str,str]])->None:
        async for db,tbl in db_tbls:
            _ = await self.get_table(db,tbl)

    async def get_table(self, database:str, table_name:str, engine):
        table_cache = self.table_cache.get((database, table_name), None)
        if table_cache is None:
            table_cache = await TableCache.new(database, table_name, engine)
        elif table_cache.expired:
            await table_cache.reset(engine)
        self.table_cache[(database,table_name)] = table_cache
        return table_cache.table

    async def run(self, query, database=None, **dbr_args):
        database = self.database if database is None else database
        engine = self.engine(database)
        if isinstance(query, str):
            query = sql_text(query)
        dbr = DataBaseResponse(query, **dbr_args) # can be overwritten
        try:
            dbr = await self.execute(engine, query, **dbr_args)
        except Exception as e:
            print("An error occurred running custom query")
            dbr.error_info = e.__repr__()
        finally:
            await engine.dispose( close = True )
        return dbr

    async def execute(self, engine, query, **dbr_args) -> DataBaseResponse:
        dbr = DataBaseResponse(query, **dbr_args)
        async with engine.connect() as connection:
            result = await connection.execute(query)
            await connection.commit()
            await dbr._receive(result)
        return dbr

    def convert_vals(self, key:str, value:object, 
                     coldata:dict, **kwargs):
        col = coldata[key]
        kwargs.update(col)
        # used for datatypes on 'write' queries
        val = col['DATA_TYPE'](value, **kwargs)
        return val.value

    def convert_header_info(self, mapp, value):
        if mapp == 'IS_NULLABLE':
            return value == 'YES'
        elif mapp == 'DATA_TYPE':
            return type_map[str(value).upper()]
        elif mapp == 'COLUMN_TYPE':
            return value.upper()
        else:
            return value
    
    async def get_header_info(self, database, table_name):
        if not database:
            database = self.database
        hdrQry = f"SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME = N'{table_name}'"
        hdrRsp = await self.run(hdrQry, database)
        res = {x['COLUMN_NAME']: {k : self.convert_header_info(k, v) 
                                  for k,v in x.items()} for x in hdrRsp}
        return res

    async def correct_types(self, database:str, table_name:str, records:list):
        coldata = await self.get_header_info(database, table_name)
        for r in records:
            for k in r:
                r[k] = self.convert_vals(k, r[k], coldata)
        return records

    def textualize(self, query):
        txt = str(query.compile(compile_kwargs={"literal_binds": True}))
        return txt

    ### SELECT, INSERT, UPDATE, DELETE, UPSERT
    async def select(self, database:str, table_name:str, params:dict=None, columns:list=None, 
               distinct:bool=False, order_by:str=None, offset:int=None, limit:int=None, 
               reverse:bool=None, textual:bool=False, response_model:object=None, 
               as_decimals:bool=False) -> DataBaseResponse | str:
        engine = self.engine(database)
        dbr = DataBaseResponse.default(database)
        try:
            table = await self.get_table(database, table_name, engine)
            if columns:
                query = select(*[table.c[col] for col in columns])
            else:
                query = select(table)
            if distinct:
                query = query.distinct()
            # Constructing complex conditions
            if params:
                query = self._construct_conditions(query, table, params)
            if order_by:
                reverse = False if reverse is None else reverse
                dir_f = asc if not reverse else desc
                query = query.order_by(dir_f(table.c[order_by]))
            if limit:
                offset = 0 if offset is None else offset
                query = query.limit(limit).offset(offset)
            if textual:
                txt = self.textualize(query)
                return txt
            dbr = await self.execute(engine, query, response_model=response_model, 
                                    as_decimals=as_decimals)
        except Exception as e:
            dbr.successful = False
            dbr.error_info = e.__repr__()
            raise e
        finally:
            await engine.dispose()
        return dbr

    async def insert(self, database:str, table_name:str,
               records:list, upsert:bool=False, 
               update_keys:list=None, textual:bool=False) -> DataBaseResponse | str:
        engine = self.engine(database)
        dbr = DataBaseResponse.default(database)
        try:
            table = await self.get_table(database, table_name, engine)
            records = await self.correct_types(database, table_name, records)
            if upsert:
                if update_keys is None:
                    update_keys = [
                        c.key for c in table.c
                    ]
                stmt = insert(table).values(records)
                update_dict = {k: stmt.inserted[k] for k in update_keys}
                query = stmt.on_duplicate_key_update(**update_dict)
            else:
                query = insert(table).values(records)
            if textual:
                txt = self.textualize(query)
                return txt
            dbr = await self.execute(engine, query)
        except Exception as e:
            dbr.successful = False
            dbr.error_info = e.__repr__()
            raise e
        finally:
            await engine.dispose()
        return dbr

    async def upsert(self, database:str, table_name:str,
               records:list, update_keys:list=None, 
               textual:bool=False):
        return await self.insert(database, table_name,
                           records, upsert=True, 
                           update_keys=update_keys,
                           textual=textual)

    async def update(self, database:str, table_name:str, 
               update:dict={}, where:dict={}, textual:bool=False):
        engine = self.engine(database)
        dbr = DataBaseResponse.default(database)
        try:
            table = await self.get_table(database, table_name, engine)
            query = sql_update(table)
            query = self._construct_conditions(query, table, where)
            query = query.values(**update)
            if textual:
                txt = self.textualize(query)
                return txt
            dbr = await self.execute(engine, query)
        except Exception as e:
            dbr.successful = False
            dbr.error_info = e.__repr__()
            raise e
        finally:
            await engine.dispose()
        return dbr
        
    async def delete(self, database:str, table_name:str,
               where:dict, textual:bool=False):
        engine = self.engine(database)
        dbr = DataBaseResponse.default(database)
        try:
            table = await self.get_table(database, table_name, engine)
            query = delete(table)
            query = self._construct_conditions(query, table, where)
            if textual:
                txt = self.textualize(query)
                return txt
            dbr = await self.execute(engine, query)
        except Exception as e:
            dbr.successful = False
            dbr.error_info = e.__repr__()
            raise e
        finally:
            await engine.dispose()
        return dbr
    
    async def clear_table(self, database:str, table_name:str, textual:bool=False):
        engine = self.engine(database)
        dbr = DataBaseResponse.default(database)
        try:
            table = await self.get_table(database, table_name, engine)
            query = delete(table)
            if textual:
                txt = self.textualize(query)
                return txt
            dbr = await self.execute(engine, query)
        except Exception as e:
            dbr.successful = False
            dbr.error_info = e.__repr__()
            raise e
        finally:
            await engine.dispose()
        return dbr

    def get_custom_builder(self, request:list[str]):
        output = [
            self.BLDRS.get(bldr, None) for bldr in request
        ]
        return output

    ### QUERY CONSTRUCTORS
    @staticmethod
    def and_(conditions):
        """Returns a callable for 'AND' condition."""
        return lambda table, query: query.where(and_(*[AsyncDataBase._process_condition(table, cond) for cond in conditions]))

    @staticmethod
    def or_(conditions):
        """Returns a callable for 'OR' condition."""
        return lambda table, query: query.where(or_(*[AsyncDataBase._process_condition(table, cond) for cond in conditions]))
    
    @staticmethod
    def not_(func, *args):
        """Returns a negated condition."""
        return lambda col: sql_not(func(*args)(col))
    
    @staticmethod
    def in_(values, _not=False):
        """Returns a callable for 'in' condition."""
        return lambda col: col.in_(values) if not _not else ~col.in_(values)

    ### QUERY FRAGMENTS
    @staticmethod
    def greater_than(value, or_equal:bool=False, _not=False):
        """Returns a callable for greater than condition."""
        def f(col):
            if or_equal:
                return col >= value if not _not else col < value
            else:
                return col > value if not _not else col <= value
        return f
    @staticmethod
    def greaterThan(value, orEqual, _not):
        return AsyncDataBase.greater_than(value, orEqual, _not)
    @staticmethod
    def less_than(value, or_equal:bool=False, _not=False):
        """Returns a callable for greater than condition."""
        def f(col):
            if or_equal:
                return col <= value if not _not else col > value
            else:
                return col < value if not _not else col >= value
        return f
    @staticmethod
    def lessThan(value, orEqual, _not):
        return AsyncDataBase.less_than(value, orEqual, _not)
    @staticmethod
    def equal_to(value, _not = False):
        return lambda col: col == value if not _not else col != value
    @staticmethod
    def equalTo(value, _not=False):
        return AsyncDataBase.equal_to(value, _not)
    @staticmethod
    def between(value1, value2, _not = False): # not inclusive
        def f(col):
            v1 = min([value1, value2])
            v2 = max([value1, value2])
            return col.between(v1, v2)
        return f
    @staticmethod
    def after(date, inclusive = False, _not = False):
        return AsyncDataBase.greater_than(date, inclusive, _not)
    @staticmethod
    def before(date, inclusive = False, _not=False):
        return AsyncDataBase.less_than(date, inclusive, _not)
    @staticmethod
    def onDay(date, _not = False):
        if isinstance(date, dt.datetime):
            date = date.date()
        return AsyncDataBase.equal_to(date, _not)
    @staticmethod
    def null(_not = False):
        return lambda col: col.is_(None) if not _not else col.isnot(None)
    @staticmethod
    def like(value, _not=False):
        return lambda col: col.like(value) if not _not else col.not_like(value)
    @staticmethod
    def starts_with(value, _not=False):
        """Returns a callable for starts with condition."""
        return AsyncDataBase.like(f"{value}%", _not)
    @staticmethod
    def startsWith(value, _not=False):
        return AsyncDataBase.starts_with(value, _not)
    @staticmethod
    def ends_with(value, _not=False):
        return AsyncDataBase.like(f"%{value}", _not)
    @staticmethod
    def endsWith(value, _not=False):
        return AsyncDataBase.ends_with(value, _not)
    @staticmethod
    def regex(value, _not=False):
        return lambda col: col.regexp_match(value) if not _not else ~col.regexp_match(value)
    @staticmethod
    def contains(value, _not=False):
        return AsyncDataBase.like(f"%{value}%", _not)
    @staticmethod
    def custom(value:str):
        lambda col: sql_text(f"`{col.table.name}`.`{col.key}` {value}")

    @staticmethod
    def _process_condition(table, condition):
        """Processes an individual condition, which can be a dict or a callable."""
        if callable(condition):
            # Condition is a callable (e.g., returned by DB.not_)
            return condition(table)
        elif isinstance(condition, dict):
            # Condition is a dictionary of key-value pairs
            return and_(*[table.c[key] == value if not callable(value) else value(table.c[key]) for key, value in condition.items()])
        else:
            raise ValueError("Invalid condition format")

    def _construct_conditions(self, query, table, params):
        """Applies conditions to the query."""
        for key, condition in params.items():
            if key in {'_AND_', '_OR_'}:
                # Assuming condition is a callable returned by DB.and_ or DB.or_
                query = condition(table, query)
            elif callable(condition):
                # Apply callable conditions
                query = query.where(condition(table.c[key]))
            else:
                # Apply simple key-value conditions
                query = query.where(table.c[key] == condition)
        return query
    
    def __repr__(self):
        return f"<DB ({self.auth.user})>"







class InvalidDate(TypeError):
    """
    Value must be a valid datetime object
    """
    pass
class NonConfiguredDataType(TypeError):
    """
    DataType not yet configured for auto-conversion
    """
    pass


# -*- coding: utf-8 -*-
import os

class Authorization:
    def __init__(self,
                 username:str,
                 password:str,
                 host:str,
                 port:int,
                 default_database:str,
                 as_async:bool=False
                 ):
        """Authentication management
        
        Example usage:
        ---
        ```
        from dbConnect import DataBase, AsyncDataBase
        
        config = {
            "username": "someuser",
            "password": "somepass",
            "host": "localhost:25252",
            "port": 3306,
            "default_database": "default_database"
        }
        
        # Instantiate the database controller
        db = DataBase(**config)
        
        # Instantiate the async version of the database controller
        adb = AsyncDataBase(**config) 
        ```
        """
        self.username        = username
        self.password        = password
        self.host            = host
        self.port            = int(port)
        self.default_database= default_database
        self.engine_name     = "asyncmy" if as_async else "pymysql"
    
    @classmethod
    def env(cls, as_async:bool=False):
        """Create Authorization from environment variables
        Required variables:
            - db_username    (username)
            - db_password    (password)
            - db_host        (host, ie localhost)
            - db_port        (port, ie 3306)
            - db_default     (default database name, ie my_db)
        """
        auth = cls(
            username = os.getenv('db_username'),
            password = os.getenv('db_password'),
            host     = os.getenv('db_host'),
            port     = os.getenv('db_port'),
            default_database = os.getenv('db_default'),
            as_async = as_async
        )
        return auth
        
    def uri(self, database:str=None):
        database = database if database is not None else self.default_database
        uri = "mysql+{engine}://{user}:{password}@{host}:{port}/{database}".format(
            engine=self.engine_name,
            user= self.username, password=self.password, host=self.host, 
            port=self.port, database=database
        )
        return uri
    
    def __repr__(self):
        return f"<Database Authorization ({self.username})>"


from setuptools import setup, find_packages

def getlong():
    with open('README.md') as f:
        data = f.read()
    return data

setup(
    name='dbmasta',
    version='0.1.1',
    author='Matt Owen',
    author_email='matt@ca-automation.app',
    description='A simple MariaDB client based on SQLAlchemy core',
    long_description=getlong(),
    long_description_content_type='text/markdown',
    url='https://github.com/mastamatto/dbConnect',
    packages=find_packages(),
    install_requires=[
        'sqlalchemy>=2.0.27',
        'asyncmy',
        'pymysql'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.7',
)
